/*

One shot learning algorithm

00/01/11
*/

match(Object, def(Class, Def)):-
	member(Conj, Def),
	satisfies(Object, Conj).
satisfies(Object, Conj):-
	\+( (member(A=VDef, Conj),
	     member(A=ODef,Object),
	     VDef \== ODef)
	  ).


learn(Class, def(Class, Def)):-
	findall(DPos, ex(Class, DPos), Pos),
	findall(DNeg, (ex(X, DNeg), X \== Class), Neg),
	learn(Pos, Neg, Def).

learn([P|Pos], Neg, [Conj|Conjs]):-
	learn_conj([P|Pos], Neg, Conj),
	remove_covered_examples([P|Pos], Conj, RemainingPos),
	learn(RemainingPos, Neg, Conjs).
learn([], _, []).


learn_conj(Pos, [N|Negs], [A=V|Conj]):-
	best_attrib_value_pair(Pos, [N|Negs], A/V),
	covered_examples([N|Negs], [A=V], CoveredNegs),
	learn_conj(Pos, CoveredNegs, Conj).
learn_conj(_, [], []).




% This Attribute Value Pair must be satisfied by at least
% one positive example and there must be at least one
% negative example that does not satisfy it

candidate(Pos, Negs, AV):-
	member(P, Pos),
	satisfies(P, [AV]),
	member(N, Negs),
	\+ satisfies(N, [AV]), !.

best_attrib_value_pair(Pos, Negs, A/V):-
	attribute_value_score(Pos, Negs, A/V/S),
	candidate(Pos, Negs, A=V),
	\+( (attribute_value_score(Pos, Negs, A1/V1/S1),
	     candidate(Pos, Negs, A1=V1),
	     S1 > S)
	  ).

attribute_value_score(Pos, Negs, A/V/S):-
	attribute_value_pair(A, V),
	score(Pos, Negs, A, V, S).

attribute_value_pair(A, V):-
	attribute(A, L),
	member(V, L).

score(Pos, Negs, A, V, S):-
	count_satisfied_examples(Pos, [A=V], NP),
	count_satisfied_examples(Negs, [A=V], NN),
	S is NP - NN.

count_satisfied_examples([Ex|Exs], Conj, N):-
	satisfies(Ex, Conj), !,
	count_satisfied_examples(Exs, Conj, N1),
	N is N1 + 1.
count_satisfied_examples([Ex|Exs], Conj, N):-
	count_satisfied_examples(Exs, Conj, N).
count_satisfied_examples([], _, 0).



% For test purposes only
negs(Class, Negs):-
	findall(ND, (ex(X, ND), X\==Class), Negs).

pos(Class, Pos):-
	findall(NP, ex(Class, NP), Pos).



/*	Attributes	*/
attribute(size, [small, large]).
attribute(shape, [long, compact, other]).
attribute(holes, [none, 1, 2, 3, many]).



/*
Examples
Copied from the book
*/

ex(nut, [size = small, shape = compact, holes = 1]).
ex(screw, [size = small, shape = long, holes = none]).
ex(key, [size = small, shape = long, holes = 1]).
ex(nut, [size = small, shape = compact, holes = 1]).
ex(key, [size = large, shape = long, holes = 1]).
ex(screw, [size = small, shape = compact, holes = none]).
ex(nut, [size = small, shape = compact, holes = 1]).
ex(pen, [size = large, shape = long, holes = none]).
ex(scissors, [size = large, shape = long, holes = 2]).
ex(pen, [size = large, shape = long, holes = none]).
ex(scissors, [size = large, shape = other, holes = 2]).
ex(key, [size = small, shape = other, holes = 2]).
